from torch import Tensor
from pege.egnn import model
from pege.utils import pdb2feats
import pandas as pd

class Pege:
    """
    Protein Enviroment Graph Embeddings

    Attributes
    ----------
    path : str
        the protein PDB file path
    coords: Tensor[1, n, 3] where n is the number of atoms
        the coordinates of the all the atoms
    feats: Tensor[1, n] where n is the number of atoms
        the features (atom classes) of all the atoms
    anumbs: list
        the atom numbers
    details: list
        the atom details
    embs: Tensor[n, 64] where n is the number of atoms
        the atom embeddings

    Methods
    -------
    get_protein(pooling=avg)
        Returns
    """

    def __init__(self, path: str):
        """
        Parameters
        ----------
        path : str
            The protein PDB file path
        """
        self.path = path
        self.coords, self.feats, self.anumbs, self.details, self.aindices = pdb2feats(path, save=False)
        self.embs = model(self.coords, self.feats).squeeze()

    def asdf(self):    
        aindices_old = list(self.aindices.keys())
        aindices_new = list(self.aindices.values())
        df_dict = {
            'anumb': self.anumbs,
            'details': self.details,
            'resnumbs': [i[1] for i in self.details],
            'embs': self.embs.detach().numpy().tolist(),
            'feats': self.feats[0],
            'coords': self.coords[0].tolist(),
            'old_anumb': [aindices_old[aindices_new.index(i)] if i in aindices_new else None for i in range(len(self.anumbs))]
        }
        return pd.DataFrame(df_dict)

    def get_protein(self, pooling: str = "avg") -> Tensor:
        """Get the protein embedding.

        If the argument pooling is set to None, all atom embeddings are returned.

        Parameters
        ----------
        pooling : str, optional
            The type of pooling to be performed (default is avg)

        Returns
        -------
        Tensor
            a tensor of (n) dimensions representing the full protein
        """
        protein_emb = self.embs
        if pooling:
            protein_emb = self.apply_pool(protein_emb, pooling)
        return protein_emb

    def get_atoms(
        self, atom_numbers: list, ignore_missing: bool = False, pooling: str = "avg"
    ) -> Tensor:
        """Get the embeddings for a list of atoms.

        If the argument pooling is set to None, all corresponding atom embeddings are returned.

        Parameters
        ----------
        atom_numbers: list
            The serial number of the atoms (in the original structure) for which to return the embeddings

        ignore_missing: bool, optional
            Controls whether an error is raised in case some atoms are not found (default is False)
        pooling : str, optional
            The type of pooling to be performed (default is avg)

        Returns
        -------
        Tensor
            a tensor of (n, 64) dimensions representing the full protein

        Raises
        ------
        Exception
            If some of the atoms present in atom_numbers are not found.
            Can be turned off by using ignore_missing.
        """
        atom_is = []
        for anumb in atom_numbers:
            if anumb not in self.aindices:
                if not ignore_missing: raise Exception(f"Atom {anumb} not present")
            else:
                a_i = self.aindices[anumb]
                atom_is.append(a_i)
        if not atom_is:
            return None
        pocket_embs = self.embs[atom_is]
        if pooling:
            pocket_embs = self.apply_pool(pocket_embs, pooling)
        return pocket_embs

    def get_residues(
        self, residue_numbers: list, ignore_missing: bool = False, pooling: str = "avg"
    ) -> Tensor:
        """Get the embeddings for a list of residues.

        If the argument pooling is set to None, all corresponding atom embeddings are returned.

        Parameters
        ----------
        residue_numbers: list
            The number of the residues (in the original structure) for which to return the embeddings

        ignore_missing: bool, optional
            Controls whether an error is raised in case some atoms are not found (default is False)
        pooling : str, optional
            The type of pooling to be performed (default is avg)

        Returns
        -------
        Tensor
            a tensor of (n, 64) dimensions representing the full protein

        Raises
        ------
        Exception
            If some of the atoms present in atom_numbers are not found.
            Can be turned off by using ignore_missing.
        """
        atom_numbers = list(self.asdf().query("resnumbs == @residue_numbers")['old_anumb'])
        return self.get_atoms(atom_numbers, ignore_missing=ignore_missing, pooling=pooling)

    @staticmethod
    def apply_pool(t: Tensor, ptype: str) -> Tensor:
        """Applies a type of pooling

        Parameters
        ----------
        t : Tensor
            the tensor to be pooled
        ptype : str
            the pooling type to be performed

        Returns
        -------
        Tensor
            the pooled tensor
        """
        if ptype == "avg":
            t = t.mean(dim=0)
        elif ptype == "sum":
            t = t.sum(dim=0)
        return t



import torch
import os
from pdbmender import mend_pdb, prepare_for_addHtaut, add_tautomers, identify_tit_sites, rm_cys_bridges, identify_cter
from pdbmender.formats import get_chains_from_file
from pege.constants import OHE_ATOMS_GRAPH, AA_HS

def read_pdb_line(line: str) -> list:
    aname = line[12:16].strip()
    anumb = int(line[5:11].strip())
    resname = line[17:21].strip()
    chain = line[21]
    resnumb = int(line[22:26])
    x = float(line[30:38])
    y = float(line[38:46])
    z = float(line[46:54])
    return (aname, anumb, resname, chain, resnumb, x, y, z)


def classify_atom(aname: str, resname: str) -> str:
    if aname.startswith('H') and (resname in AA_HS and aname in AA_HS[resname]):
        return 'H'

    if aname in ("N", "O"):
        if resname != "NTR":
            return aname
        else:
            return "NZ_LYS"
    elif resname in ("CTR", "ASP", "GLU"):
        return "O_COOH"
    elif resname in ("GLN", "ASN"):
        return f"{aname[0]}_AMIDE"
    elif resname == "ARG" and aname in ("NH1", "NH2"):
        return "NH_ARG"
    elif resname == "CY0" and aname == "SG":
        return "SG_CYS"
    else:
        return f"{aname}_{resname}"

def iter_f(fname):
    with open(fname) as f:
        for line in f:
            if line.startswith('ATOM '):
                yield line


def get_atom_info(fname):
    original_atom_details = {}
    termini_restypes = {}
    for line in iter_f(fname):
        aname, anumb, resname, chain, resnumb, x, y, z = read_pdb_line(line)
        inscode = line[26]
        details = (chain, resnumb, resname, aname, inscode)
        original_atom_details[anumb] = details
        
    original_atom_numbs = original_atom_details.keys()
    original_atom_infos = original_atom_details.values()
    
    return original_atom_numbs, original_atom_infos


def fix_structure(fname, chains_res, ff):
    sysname = fname.split('.pdb')[0]
    pdb_cleaned = f"{sysname}_cleaned.pdb"
    logfile_mend = "LOG_pdb2pqr"
    mend_pdb(fname, pdb_cleaned, ff, ff, logfile=logfile_mend)

    chains_res, cys_bridges = rm_cys_bridges(chains_res, logfile_mend)

    old_ctrs = {chain: resnumb for chain, residues in chains_res.items() for resnumb, resname in residues.items() if resname == 'CTR'}
    new_ctrs = identify_cter(pdb_cleaned, old_ctrs)
    for chain, resnumb in new_ctrs.items():
        chains_res[chain][str(resnumb)] = "CTR"

    output_pdb = f"{sysname}_final.pdb"
    _, renumbered = add_tautomers(pdb_cleaned, chains_res, ff, output_pdb)
    return output_pdb, chains_res, renumbered



def get_termini_info(fname, chains_res):
    termini_resnumbs = {chain: [None, None] for chain in chains_res.keys()}
    for chain, residues in chains_res.items():
        for resnumb, resname in residues.items():
            resnumb = int(resnumb)
            if resname == "NTR" and not termini_resnumbs[chain][0]:
                termini_resnumbs[chain][0] = resnumb
            elif resname == "CTR" and not termini_resnumbs[chain][1]:
                termini_resnumbs[chain][1] = resnumb


    termini_resnames = {chain: [None, None] for chain in chains_res.keys()}
    for line in iter_f(fname):
        aname, anumb, resname, chain, resnumb, x, y, z = read_pdb_line(line)

        if termini_resnumbs[chain][0] and resnumb == termini_resnumbs[chain][0]:
            termini_resnames[chain][0] = resname
        elif termini_resnumbs[chain][1] and resnumb == termini_resnumbs[chain][1]:
            termini_resnames[chain][1] = resname

    return termini_resnumbs, termini_resnames


def pdb2feats(fname: str, save=True, ff="GROMOS") -> tuple:
    chains = get_chains_from_file(fname)
    chains_res = identify_tit_sites(fname, chains, add_ser_thr=True)

    original_atom_details = get_atom_info(fname)

    output_pdb, chains_res, renumbered = fix_structure(fname, chains_res, ff)

    termini_details = get_termini_info(fname, chains_res)

    coords, feats, anumbs, details, aindices = encode_structure(output_pdb, termini_details, original_atom_details, renumbered)

    if not save:
        os.system(f'rm -f {output_pdb} *_cleaned.pdb* LOG* removed.pqr')

    return coords, feats, anumbs, details, aindices

def encode_structure(output_pdb, termini_details, original_atom_details, renumbered):
    termini_resnumbs, termini_resnames = termini_details
    original_atom_numbs, original_atom_infos = original_atom_details
    
    coords, feats, anumbs, details = [], [], [], []
    aindices = {}
    
    for line in iter_f(output_pdb):
        aname, anumb, resname, chain, resnumb, x, y, z = read_pdb_line(line)

        b, icode = line[16], line[26]
        if b not in (" ", "A") or icode != " ":
            continue

        if chain in termini_resnumbs and resnumb in termini_resnumbs[chain]:
            i_ter = termini_resnumbs[chain].index(resnumb)
            ter_type = "NTR" if i_ter == 0 else "CTR"

            if ter_type == "NTR" and aname in ('N', 'H1', 'H2', 'H3'):
                resname = "NTR"
            elif aname in ('O1', 'O2', 'HO11', 'HO21', 'HO12', 'HO22'):
                resname = "CTR"

        if aname[0] not in "NOS" and (resname not in AA_HS or aname not in AA_HS[resname]):
            continue
            
        res = classify_atom(aname, resname)

        if res in OHE_ATOMS_GRAPH:
            res_ohe = OHE_ATOMS_GRAPH.index(res)

            ainfo = (chain, resnumb, resname, aname, ' ')

            coords.append((x, y, z))
            feats.append(res_ohe)
            anumbs.append(anumb)
            details.append(ainfo)

            if resnumb in renumbered.keys():
                old_resnumb, inscode = renumbered[resnumb]
                ainfo_search = (chain, old_resnumb, resname, aname, inscode)
            elif resname in ('NTR', 'CTR'):
                old_aname = aname
                if resname == 'NTR':
                    old_resname = termini_resnames[chain][0]
                elif resname == 'CTR':
                    converted_atoms = {'O1': 'O', 'O2': 'OXT'}
                    if aname in converted_atoms:
                        old_aname = converted_atoms[aname]
                    old_resname = termini_resnames[chain][1]
                ainfo_search = (chain, resnumb, old_resname, old_aname, ' ')
            else:
                ainfo_search = ainfo

            if ainfo_search in original_atom_infos:
                old_anumb = list(original_atom_numbs)[list(original_atom_infos).index(ainfo_search)]
                aindices[old_anumb] = len(anumbs) - 1
                # print(f'{line.strip()}   {res:10}   {old_anumb}')
            
            elif res != 'H':
                warning_msg = f'WARNING: atom {anumb} in {output_pdb} will be ignored'
                print(warning_msg)

    coords = torch.tensor(coords).reshape(1, -1, 3)
    feats = torch.tensor(feats).reshape(1, -1)

    return coords, feats, anumbs, details, aindices




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pege']

package_data = \
{'': ['*']}

install_requires = \
['egnn-pytorch>=0.2.6,<0.3.0',
 'pdbmender>=0.3.0,<0.4.0',
 'pytorch-lightning>=1.4.7,<2.0.0',
 'torch>=1.9.0,<2.0.0']

setup_kwargs = {
    'name': 'pege',
    'version': '0.1.0',
    'description': 'Protein embeddings to describe local electrostic enviroments',
    'long_description': '# Protein Enviroment Graph Embeddings (PEGE)\n\nProtein embeddings to describe local electrostic enviroments\n\n# Installation & Basic Usage\n\nPEG is installable from the Pypi repo:\n```bash\npython3 -m pip install pege\n```\n\nPege can be used to obtain protein embeddings as well as descriptors for specific `atom_numbers` from a `pdb` file:\n```python\nfrom pege import Pege\n\nprotein = Pege(<pdb>)\nprotein_emb = protein.get_protein()\natoms_emb = protein.get_atoms([<atom_numbers>])\n```\n\n# Documentation\nTBA\n\n# License\nThis source code is licensed under the MIT license found in the LICENSE file in the root directory of this source tree.\n\n# Contacts\nPlease submit a github issue to report bugs and to request new features. Alternatively, you may email the developer [directly](mailto:pdreis@fc.ul.pt).\n\n',
    'author': 'Pedro Reis',
    'author_email': 'pdreis@fc.ul.pt',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bayer-science-for-a-better-life/pege',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)
